with Ada.Text_IO;
use Ada.Text_IO;
with Ada.Strings.Fixed;
with Ada.Strings.Maps;

package body QUCS.Datasets.IO is

   type Delimiter_Descriptor is
      record
         Open : Boolean;
         Parameters : Name_Lists.Vector;
      end record;

   procedure Dump (X : Delimiter_Descriptor) is
   begin
      Put_Line (Standard_Error, "Open : " & Boolean'Image (X.Open));
      for I in 1 .. X.Parameters.Last_Index loop
         Put_Line (Standard_Error,
                   "[" & Integer'Image (I) & "]"
                   &
                   "'" & X.Parameters.Element (I) & "'");
      end loop;
   end Dump;

   function Parse_Delimiter (Input     : File_Type)
                             return  Delimiter_Descriptor
   is
      Line   : String := Get_Line (Input);
      Cursor : Positive := Line'First;

      Result : Delimiter_Descriptor;

      procedure Raise_Parsing_Error (Msg : String) is
      begin
         raise Parsing_Error with Msg;
      end Raise_Parsing_Error;

      procedure Expect (Item : Character;
                        Msg  : String) is
      begin
--           Put_Line (Cursor'Img & Line'Last'Img);
--           Put_Line (" '" & Line (Cursor) & Item & "'");

         if Cursor > Line'Last then
            raise Constraint_Error;
         end if;

         if Line (Cursor) /= Item then
            raise Program_Error;
         end if;

         if Cursor > Line'Last or else Line (Cursor) /= Item then
            Raise_Parsing_Error (Msg);
         end if;

         Cursor := Cursor + 1;
      end Expect;

      procedure Expect_EOL (Msg  : String) is
      begin
         if Cursor <= Line'Last then
            Raise_Parsing_Error (Msg);
         end if;
      end Expect_EOL;

      function Is_Next (Item : Character) return Boolean is
      begin
         if Cursor <= Line'Last and then  Line (Cursor) = Item then
            Cursor := Cursor + 1;
            return True;
         else
            return False;
         end if;
      end Is_Next;

      function Next_Token return String is
         use Ada.Strings.Fixed;
         use Ada.Strings.Maps;

         First : Positive;
         Last : Natural;
      begin
         Find_Token (Source => Line,
                     Set    => To_Set(" >"),
                     From   => Cursor,
                     Test   => Ada.Strings.Outside,
                     First  => First,
                     Last   => Last);

         if Last = 0 then
            return "";
         else
            Cursor := Last + 1;
            return Line (First .. Last);
         end if;
      end Next_Token;
   begin
--        Put_Line ("Using '" & Line & "'");

      Expect ('<', "Delimiter line does not begin with '<'");

      if Is_Next ('/') then
         Result.Open := False;
      else
         Result.Open := True;
      end if;

      loop
         declare
            Token : String := Next_Token;
         begin
            exit when Token = "";
            Result.Parameters.Append (Token);
         end;
      end loop;

      Expect ('>',  "Delimiter line does not end with '>'");

      Expect_EOL ("Garbage after end of delimiter");

      return Result;
   end Parse_Delimiter;

   --------------------
   -- Expect_Closing --
   --------------------

   procedure Expect_Closing (Input : File_Type;
                             Label : String) is
      D : Delimiter_Descriptor := Parse_Delimiter (Input);
   begin
      if D.Open then
         raise Program_Error
           with "Expected closing delimiter, found an open one";
      end if;


      if D.Parameters.Element (1) /= Label then
         raise Parsing_Error
           with "Expected closing '" & Label & "', found '"
             & D.Parameters.Element (1) & "'";
      end if;
   end Expect_Closing;

   -------------------
   -- Load_Variable --
   -------------------

   procedure Load_Variable (Container : in out Dataset;
                            Input     : File_Type;
                            Name      : String)
   is
      N_Item    : Positive := Container.Get (Name).Total_Size;
   begin
      for I in 1 .. N_Item loop
         Container.Set (Name  => Name,
                        Index => I,
                        Value => Float'Value (Get_Line (Input)));
      end loop;
   end Load_Variable;


   ----------
   -- Load --
   ----------

   function Load (Filename : String) return Dataset is
      use Ada.Containers;

      -------------------------------
      -- Load_Independent_Variable --
      -------------------------------

      procedure Load_Independent_Variable (Container  : in out Dataset;
                                           Input      : File_Type;
                                           Name       : String;
                                           Parameters : Name_Lists.Vector)
      is
         N_Item : Positive;
      begin
         if Parameters.Length /= 3 then
            raise Parsing_Error with "Bad open delimiter line";
         end if;

         N_Item := Natural'Value (Parameters.Element (3));

         Container.Add_Variable (Name, N_Item);

         Load_Variable (Container => Container,
                        Input     => Input,
                        Name      => Name);
      end Load_Independent_Variable;

      procedure Load_Dependent_Variable (Container  : in out Dataset;
                                         Input      : File_Type;
                                         Name       : String;
                                         Parameters : Name_Lists.Vector)
      is
         Size : Variables.Dim_Array (3 .. Parameters.Last_Index);
      begin
         for I in Size'Range loop
            declare
               Indep_Var : String := Parameters.Element (I);
            begin
               if not Container.Exist (Indep_Var) then
                  raise Parsing_Error
                    with "Unknown independent variable '"
                      & Indep_Var & "'";
               end if;

               Size (I) := Container.Get (Indep_Var).Total_Size;
            end;
         end loop;

         Container.Add_Variable (Name, Size);

         for I in 3 .. Parameters.Last_Index loop
            Container.Add_Dependency
              (Name       => Name,
               Depends_On => Parameters.Element (I));
         end loop;

         Load_Variable (Container => Container,
                        Input     => Input,
                        Name      => Name);
      end Load_Dependent_Variable;


      Input : File_Type;
      Result : Dataset;
      Delim : Delimiter_Descriptor;
   begin
      Open (File     => Input,
            Mode     => Ada.Text_IO.In_File,
            Name     => Filename);

      Delim := Parse_Delimiter (Input);


      if not
        (Delim.Open and
           Delim.Parameters.Element (1) = "Qucs" and
           Delim.Parameters.Element (2) = "Dataset") then
         Dump (Delim);

         raise Parsing_Error with "Wrong header";
      end if;

      while not End_Of_File (Input) loop
         Delim := Parse_Delimiter (Input);

         if not Delim.Open then
            Dump (Delim);

            raise Parsing_Error with "Open delimiter expected";
         end if;

         if Delim.Parameters.Length < 3 then
            Dump (Delim);

            raise Parsing_Error with "Bad open delimiter line";
         end if;

         declare
            Delim_Type : String := Delim.Parameters.Element (1);
            Name       : String := Delim.Parameters.Element (2);
         begin
            if Delim_Type = "indep" then

               Load_Independent_Variable (Result, Input, Name, Delim.Parameters);

               Expect_Closing (Input, "indep");
            elsif Delim_Type = "dep" then

               Load_Dependent_Variable (Result, Input, Name, Delim.Parameters);

               Expect_Closing (Input, "dep");
            else
               raise Parsing_Error with "Unrecognized delimiter type";
            end if;
         end;
      end loop;

      Close (Input);

      return Result;
   end Load;

end QUCS.Datasets.IO;
